import struct

import numpy as np
from ..log import logger


def readma2(fname):
    """A function for reading binary map files (``*.ma2``) for nek5000.

    The map file comtains, for each element in the mesh, the id of the MPI rank that owns it
    followed by the ids of the vertices of the element in the global address space.

    The partitioning is determined by generating the undirected graph formed by the mesh, then
    repeatedly computing and deviding the graph using the Fiedler vector of the graph Laplacian.

    Parameters
    ----------
    fname : str
            file name

    Returns
    -------
    cell  : 2d int ndarray
            list of the vertices for each element of the mesh (in global address space)
    procmap  : 1d int ndarray
            processor map (0-based) indicating ownership for each element of the mesh
    """
    try:
        infile = open(fname, "rb")
    except OSError as e:
        logger.critical(f"I/O error ({e.errno}): {e.strerror}")
        return -1
    #
    # read header
    header = infile.read(132).split()
    nel = int(header[1])

    # number of active elements (nrank - noutflow)
    # nactive = int(header[2])

    # total number of element vertices in the mesh
    # npts  = (2**ldim)*nel
    npts = int(header[5])

    # number of unique element vertices in the mesh
    # nrank = int(header[6])

    # number of points on outflow boundaries ('o  ')
    # noutflow = int(header[7])

    # NOTE: these values can be computed from the others but are included in the
    # header
    # number of levels in the binary partition tree
    # depth = log2(nel)
    # depth = int(header[3])

    # maximum number of elements in partition tree of depth d
    # d2    = 2**d
    # d2 = int(header[4])
    # always double precision
    wdsz = 4
    inttype = "i"

    # detect endianness
    etagb = infile.read(4)
    etagL = struct.unpack("<f", etagb)[0]
    etagL = int(etagL * 1e5) / 1e5
    etagB = struct.unpack(">f", etagb)[0]
    etagB = int(etagB * 1e5) / 1e5
    if etagL == 6.54321:
        logger.debug("Reading little-endian file\n")
        emode = "<"
        # endian = "little"
    elif etagB == 6.54321:
        logger.debug("Reading big-endian file\n")
        emode = ">"
        # endian = "big"
    else:
        logger.error("Could not interpret endianness")
        return -3

    # read the entire contents of the file
    # for each element, there are nvert vertices and a processor id
    nvert = int(npts / nel)  # 2**ldim
    buf = infile.read((nvert + 1) * wdsz * nel)

    # processor map (0-based)
    procmap = np.empty((nel,))
    # list of vertices for each element (in global address space)
    cell = np.empty((nel, nvert))
    for iel in range(nel):
        fi = np.frombuffer(
            buf,
            dtype=emode + inttype,
            count=nvert + 1,
            offset=(nvert + 1) * wdsz * iel,
        )
        procmap[iel] = fi[0]
        cell[iel, :] = fi[1:]
    # close file
    infile.close()
    #
    # output
    return cell, procmap



"""Core data structures for pymech"""
import copy
from textwrap import dedent, indent
import itertools
from itertools import product
from functools import reduce, partial

import numpy as np
from pymech.log import logger


"""Repeat N times. Pythonic idiom to use when the iterated value is discarded.

Example
-------
Instead of:

>>> [0 for _ in range(10)]

You could use:

>>> [0 for _ in repeat(10)]

"""
repeat = partial(itertools.repeat, None)


# ==============================================================================
class DataLims:
    """A class containing the extrema of all quantities stored in the mesh

    Attributes
    ----------
    - pos:  x,y,z   min,max
    - vel:  u,v,w   min,max
    - pres: p       min,max
    - temp: T       min,max
    - scal: s_i     min,max

    """

    def __init__(self, elements):
        self._variables = ("pos", "vel", "pres", "temp", "scal")

        aggregated_lims = reduce(self._lims_aggregator, elements)
        for var in self._variables:
            agg_lims_var = aggregated_lims[var]
            # set minimum, maximum of variables as a nested tuple
            setattr(self, var, tuple(zip(*agg_lims_var)))

        # prevent further mutation of attributes via __setattr__
        self._initialized = True

    def __repr__(self):
        return dedent(
            f"""\
          * x:         {self.pos[0]}
          * y:         {self.pos[1]}
          * z:         {self.pos[2]}"""
        )

    def __setattr__(self, name, value):
        if hasattr(self, "_initialized") and self._initialized:
            raise AttributeError(f"Setting attribute {name} is not permitted")
        else:
            super().__setattr__(name, value)

    def _lims_per_element(self, elem):
        """Get local limits for a given element."""
        if isinstance(elem, dict):
            return elem

        axis = (1, 2, 3)
        elem_lims = {
            var: (getattr(elem, var).min(axis), getattr(elem, var).max(axis))
            for var in self._variables
        }
        return elem_lims

    def _lims_aggregator(self, elem1, elem2):
        """Reduce local limits to global limits."""
        l1 = self._lims_per_element(elem1)
        l2 = self._lims_per_element(elem2)

        aggregated_lims = {
            var: (
                np.minimum(l1[var][0], l2[var][0]),
                np.maximum(l1[var][1], l2[var][1]),
            )
            for var in self._variables
        }
        return aggregated_lims


# ==============================================================================
class Elem:
    """A class containing one hexahedral element of Nek5000/SIMSON flow
    field.

    Parameters
    ----------
    var : iterable
        Iterable of integers of size 5, indicating how many variables are to be initialized
    lr1 : iterable
        Iterable of integers of size 3, defining the shape of an element as ``(lx, ly, lz)``
    nbc : int
        Number of boundary conditions
    dtype : str
        Floating point data type. Typical values are 'f4' or 'float32' for
        single precision, 'f8' or 'float64' for double precision

    """

    def __init__(self, var, lr1, nbc, dtype="float64"):
        #                    x,y,z   lz      ly      lx
        self.pos = np.zeros((3, lr1[2], lr1[1], lr1[0]), dtype=dtype)
        #                    one per edge
        self.curv = np.zeros((12, 5), dtype=dtype)
        #             curvature type
        self.ccurv = ["" for _ in repeat(12)]
        #                    u,v,w   lz      ly      lx
        self.vel = np.zeros((3, lr1[2], lr1[1], lr1[0]), dtype=dtype)
        #                    p       lz      ly      lx
        self.pres = np.zeros((var[2], lr1[2], lr1[1], lr1[0]), dtype=dtype)
        #                    T       lz      ly      lx
        self.temp = np.zeros((var[3], lr1[2], lr1[1], lr1[0]), dtype=dtype)
        #                    s_i     lz      ly      lx
        self.scal = np.zeros((var[4], lr1[2], lr1[1], lr1[0]), dtype=dtype)
        #                    list of 8 parameters, one per face
        #                    one column for velocity, one for temperature, and one for each scalar
        self.bcs = np.zeros((nbc, 6), dtype="U3, i4, i4" + f", {dtype}" * 5)

    def __repr__(self):
        message = f"<elem centered at {self.centroid}>"
        return message

    @property
    def centroid(self):
        return self.pos.mean(axis=(1, 2, 3))

    def smallest_edge(self):
        """returns the length of the smallest edge, neglecting curvature"""

        # get coordinates of points
        x1, y1, z1 = self.pos[:, 0, 0, 0]
        x2, y2, z2 = self.pos[:, 0, 0, 1]
        x3, y3, z3 = self.pos[:, 0, 1, 0]
        x4, y4, z4 = self.pos[:, 0, 1, 1]
        # compute squares of edges lengths
        edges_l2 = np.zeros((12,))
        edges_l2[0] = (x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2
        edges_l2[1] = (x3 - x2) ** 2 + (y3 - y2) ** 2 + (z3 - z2) ** 2
        edges_l2[2] = (x4 - x3) ** 2 + (y4 - y3) ** 2 + (z4 - z3) ** 2
        edges_l2[3] = (x1 - x4) ** 2 + (y1 - y4) ** 2 + (z1 - z4) ** 2
        # the dimension is not stored but we can cheat
        if self.pos.shape[1] > 1:
            ndim = 3
        else:
            ndim = 2
        if ndim > 2:
            # in 3D, do the same for the upper face, and also the side edges
            x5, y5, z5 = self.pos[:, 1, 0, 0]
            x6, y6, z6 = self.pos[:, 1, 0, 1]
            x7, y7, z7 = self.pos[:, 1, 1, 0]
            x8, y8, z8 = self.pos[:, 1, 1, 1]
            edges_l2[4] = (x6 - x5) ** 2 + (y6 - y5) ** 2 + (z6 - z5) ** 2
            edges_l2[5] = (x7 - x6) ** 2 + (y7 - y6) ** 2 + (z7 - z6) ** 2
            edges_l2[6] = (x8 - x7) ** 2 + (y8 - y7) ** 2 + (z8 - z7) ** 2
            edges_l2[7] = (x5 - x8) ** 2 + (y5 - y8) ** 2 + (z5 - z8) ** 2
            edges_l2[8] = (x5 - x1) ** 2 + (y5 - y1) ** 2 + (z5 - z1) ** 2
            edges_l2[9] = (x6 - x2) ** 2 + (y6 - y2) ** 2 + (z6 - z2) ** 2
            edges_l2[10] = (x7 - x3) ** 2 + (y7 - y3) ** 2 + (z7 - z3) ** 2
            edges_l2[11] = (x8 - x4) ** 2 + (y8 - y4) ** 2 + (z8 - z4) ** 2
            return np.sqrt(edges_l2.min())
        else:
            return np.sqrt(edges_l2[:4].min())

    def face_center(self, i):
        """Return the coordinates (x, y, z) of the center of the face number i"""

        if i == 0:
            kx1, ky1, kz1 = 0, 0, 0
            kx2, ky2, kz2 = 0, 0, -1
            kx3, ky3, kz3 = -1, 0, 0
            kx4, ky4, kz4 = -1, 0, -1
        elif i == 1:
            kx1, ky1, kz1 = -1, 0, 0
            kx2, ky2, kz2 = -1, 0, -1
            kx3, ky3, kz3 = -1, -1, 0
            kx4, ky4, kz4 = -1, -1, -1
        elif i == 2:
            kx1, ky1, kz1 = 0, -1, 0
            kx2, ky2, kz2 = 0, -1, -1
            kx3, ky3, kz3 = -1, -1, 0
            kx4, ky4, kz4 = -1, -1, -1
        elif i == 3:
            kx1, ky1, kz1 = 0, 0, 0
            kx2, ky2, kz2 = 0, 0, -1
            kx3, ky3, kz3 = 0, -1, 0
            kx4, ky4, kz4 = 0, -1, -1
        elif i == 4:
            kx1, ky1, kz1 = 0, 0, 0
            kx2, ky2, kz2 = 0, -1, 0
            kx3, ky3, kz3 = -1, 0, 0
            kx4, ky4, kz4 = -1, -1, 0
        elif i == 5:
            kx1, ky1, kz1 = 0, 0, -1
            kx2, ky2, kz2 = 0, -1, -1
            kx3, ky3, kz3 = -1, 0, -1
            kx4, ky4, kz4 = -1, -1, -1
        else:
            logger.error(f"Invalid face number {i} (must be between 0 and 5)")
        (x1, y1, z1) = self.pos[:, kz1, ky1, kx1]
        (x2, y2, z2) = self.pos[:, kz2, ky2, kx2]
        (x3, y3, z3) = self.pos[:, kz3, ky3, kx3]
        (x4, y4, z4) = self.pos[:, kz4, ky4, kx4]
        return (
            0.25 * (x1 + x2 + x3 + x4),
            0.25 * (y1 + y2 + y3 + y4),
            0.25 * (z1 + z2 + z3 + z4),
        )


# ==============================================================================
class HexaData:
    """A class containing data related to a hexahedral mesh"""

    def __init__(self, ndim, nel, lr1, var, nbc=0, dtype="float64"):
        self.ndim = ndim
        self.nel = nel
        self.ncurv = []
        self.nbc = nbc
        self.var = var
        self.lr1 = lr1
        self.time = []
        self.istep = []
        self.wdsz = []
        self.endian = []
        if isinstance(dtype, type):
            # For example np.float64 -> "float64"
            dtype = dtype.__name__

        self.elem = [Elem(var, lr1, nbc, dtype) for _ in repeat(nel)]
        self.elmap = np.linspace(1, nel, nel, dtype=np.int32)

    def __repr__(self):
        representation = dedent(
            f"""\
        <pymech.core.HexaData>
        Dimensions:    {self.ndim}
        Precision:     {self.wdsz} bytes
        Mesh limits:\n{indent(repr(self.lims), " "*10)}
        Time:
          * time:      {self.time}
          * istep:     {self.istep}
        Elements:
          * nel:       {self.nel}
          * elem:      [{self.elem[0]}
                        ...
                        {self.elem[-1]}]
        """
        )

        return representation

    @property
    def lims(self):
        return DataLims(self.elem)

    def check_connectivity(self, tol=1e-3):
        """Check element connectivity, specifically for matching boundary
        conditions and geometry. Errors are reported as logging messages.

        Parameters
        ----------
        tol : float
            relative tolerance (compared to the smallest edge of adjacent elements)
            for detecting whether faces are at the same location

        """
        dim = self.ndim
        err = False
        for (iel, el), ibc, iface in product(
            enumerate(self.elem), range(self.nbc), range(2 * dim)
        ):
            cbc = el.bcs[ibc, iface][0]
            if cbc == "E" or cbc == "P":
                connected_iel = int(el.bcs[ibc, iface][3]) - 1
                connected_face = int(el.bcs[ibc, iface][4]) - 1
                xc, yc, zc = el.face_center(iface)
                if connected_iel < 0 or connected_iel >= self.nel:
                    err = True
                    logger.error(
                        f"face {iface} of element {iel} is connected ('{cbc}') to face "
                        f"{connected_face} of the nonexistent element {connected_iel}"
                    )
                    logger.error(f"face center: ({xc:.6e} {yc:.6e} {zc:.6e})")
                else:
                    cbc1 = self.elem[connected_iel].bcs[ibc, connected_face][0]
                    iel1 = int(self.elem[connected_iel].bcs[ibc, connected_face][3]) - 1
                    iface1 = (
                        int(self.elem[connected_iel].bcs[ibc, connected_face][4]) - 1
                    )
                    xc1, yc1, zc1 = self.elem[connected_iel].face_center(connected_face)
                    if cbc1 != cbc or iel1 != iel or iface1 != iface:
                        err = True
                        logger.error(
                            "mismatched boundary conditions: "
                            f"face {iface + 1} of element {iel + 1} with "
                            f"condition '{cbc}' is connected to face {connected_face + 1} "
                            f"of element {connected_iel + 1}, which has condition '{cbc1}' "
                            f"and points to face {iface1} of element {iel1}"
                        )
                        logger.error(
                            f"face centers: ({xc:.6e} {yc:.6e} {zc:.6e}), ({xc1:.6e} {yc1:.6e} {zc1:.6e})"
                        )
                    elif cbc == "E":  # no check for 'P' yet, but it should be possible
                        dist = np.sqrt(
                            (xc - xc1) ** 2 + (yc - yc1) ** 2 + (zc - zc1) ** 2
                        )
                        max_dist = tol * min(
                            el.smallest_edge(), self.elem[connected_iel].smallest_edge()
                        )
                        if dist > max_dist:
                            err = True
                            logger.error(
                                "mismatched face locations: "
                                f"face {iface + 1} of element {iel + 1} and "
                                f"face {connected_face + 1} of element {connected_iel + 1} "
                                "are connected but are not at the same location"
                            )
                            logger.error(
                                f"face centers: ({xc:.6e} {yc:.6e} {zc:.6e}), ({xc1:.6e} {yc1:.6e} {zc1:.6e})"
                            )
        return not err

    def check_bcs_present(self):
        """
        Returns True if and only if all faces of all elements have boundary conditions applied.

        Note that this function returning False does not mean the mesh is invalid: it is not mandatory
        to define internal boundary conditions for Nek5000.
        """

        res = True
        for (iel, el), ibc, iface in product(
            enumerate(self.elem), range(self.nbc), range(2 * self.ndim)
        ):
            if el.bcs[ibc, iface][0] == "":
                res = False
                logger.error(
                    f"missing boundary condition at element {iel}, face {iface}, field {ibc}"
                )
        return res

    def merge(self, other, tol=1e-2, ignore_empty=True, ignore_all_bcs=False):
        """
        Merges another :class:`pymech.core.HexaData` into the current one and connects it

        Parameters
        ----------
        other: :class:`pymech.core.HexaData`
                mesh to merge into self

        tol: float
                maximum distance, relative to the smallest edge of neighbouring elements, at which faces are considered touching

        ignore_empty: bool
                if True, the faces with an empty boundary condition ('') will be treated as internal faces and will not be merged.
                This is useful if internal boundary conditions are not defined and will make the operation *much* faster,
                but requires boundary conditions to be defined on the faces to be merged.

        ignore_all_bcs: bool
                if True, the boundary conditions will not be changed.
                This is likely to result in invalid boundary conditions at the interface between the merged meshes.
                This option is intended for fast merging in a situation in which the boundary conditions
                are either irrelevant or will be defined or corrected later.
        """

        # perform some consistency checks
        if self.ndim != other.ndim:
            logger.error(
                f"Cannot merge meshes of dimensions {self.ndim} and {other.ndim}!"
            )
            return -1
        if self.lr1[0] != other.lr1[0]:
            logger.error(
                "Cannot merge meshes of different polynomial orders ({} != {})".format(
                    self.lr1[0], other.lr1[0]
                )
            )
            return -2

        # add the new elements (in an inconsistent state if there are internal boundary conditions)
        nel1 = self.nel
        self.nel = self.nel + other.nel
        self.ncurv = self.ncurv + other.ncurv
        self.elmap = np.concatenate((self.elmap, other.elmap + nel1))
        # the deep copy is required here to avoid leaving the 'other' mesh in an inconsistent state by modifying its elements
        mesh_add = copy.deepcopy(other)
        mesh_add.offset_connectivity(nel1)
        self.elem = self.elem + mesh_add.elem

        # check how many boundary condition fields we have
        nbc = min(self.nbc, other.nbc)

        # glue common faces together
        # only look for the neighbour in the first BC field because it should be the same in all fields.
        nfaces = 2 * self.ndim
        nchanges = 0  # counter for the boundary conditions connected
        if nbc == 0 or ignore_all_bcs:
            # Quickly exit the function
            logger.debug("no pairs of faces to merge")
            return nchanges

        for iel0, iface0 in product(range(nel1, self.nel), range(nfaces)):
            elem0 = self.elem[iel0]
            bc0 = elem0.bcs[0, iface0][0]

            if bc0 != "E" and not (ignore_empty and bc0 == ""):
                # boundary element, look if it can be connected to something
                for iel1, iface1 in product(range(nel1), range(nfaces)):
                    elem1 = self.elem[iel1]
                    bc1 = elem1.bcs[0, iface1][0]

                    if bc1 != "E" and not (ignore_empty and bc1 == ""):
                        # if the centers of the faces are close, connect them together
                        x0, y0, z0 = elem0.face_center(iface0)
                        x1, y1, z1 = elem1.face_center(iface1)
                        dist = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2 + (z1 - z0) ** 2)
                        dist_ref = min(elem0.smallest_edge(), elem1.smallest_edge())
                        if dist <= tol * dist_ref:
                            # reconnect the periodic faces together (assumes that all fields are periodic)
                            if bc0 == bc1 == "P":
                                iel_p0 = int(elem0.bcs[0, iface0][3]) - 1
                                iel_p1 = int(elem1.bcs[0, iface1][3]) - 1
                                iface_p0 = int(elem0.bcs[0, iface0][4]) - 1
                                iface_p1 = int(elem1.bcs[0, iface1][4]) - 1
                                for ibc in range(nbc):
                                    elem_p0_bcs = self.elem[iel_p0].bcs[ibc, iface_p0]
                                    elem_p1_bcs = self.elem[iel_p1].bcs[ibc, iface_p1]

                                    elem_p0_bcs[0] = "P"
                                    elem_p1_bcs[0] = "P"
                                    elem_p0_bcs[3] = iel_p1 + 1
                                    elem_p1_bcs[3] = iel_p0 + 1
                                    elem_p0_bcs[4] = iface_p1 + 1
                                    elem_p1_bcs[4] = iface_p0 + 1

                            for ibc in range(nbc):
                                elem0.bcs[ibc, iface0][0] = "E"
                                elem1.bcs[ibc, iface1][0] = "E"
                                elem0.bcs[ibc, iface0][3] = iel1 + 1
                                elem1.bcs[ibc, iface1][3] = iel0 + 1
                                elem0.bcs[ibc, iface0][4] = iface1 + 1
                                elem1.bcs[ibc, iface1][4] = iface0 + 1
                            nchanges = nchanges + 1

        logger.debug(f"merged {nchanges} pairs of faces")
        return nchanges

    def get_points(self):
        """
        Returns an array containing the coordinates of all the points in the mesh as a (nel, lx1*ly1*lz1, 3) array
        """

        lx1, ly1, lz1 = self.lr1
        nxyz = lx1 * ly1 * lz1
        xyz = np.zeros((self.nel, nxyz, 3))
        for el, lxyz in zip(self.elem, xyz):
            for idim in range(3):
                lxyz[:, idim] = el.pos[idim, ...].ravel()

        return xyz

    def update_ncurv(self):
        """
        Updates the metadata `ncurv` integer to match the actual number of curved faces present in the mesh
        """

        ncurv = 0
        for el in self.elem:
            for iedge in range(12):
                if el.ccurv[iedge] != "":
                    ncurv = ncurv + 1
        self.ncurv = ncurv

    def offset_connectivity(self, offset: int, iel_min=0):
        """
        Adds a value to the index of the elements connected via internal or periodic
        boundary conditions to elements of the mesh. This is used to keep the connectivity
        valid when deleting or inserting elements in the mesh.

        Parameters
        ----------
        offset  : int
            The value by which to offset the indices
        iel_min : int
            The first element (in zero-based indexing) to offset
        """

        for el, ibc, iface in product(self.elem, range(self.nbc), range(2 * self.ndim)):
            bc = el.bcs[ibc, iface][0]
            if bc == "E" or bc == "P":
                if (
                    int(el.bcs[ibc, iface][3]) > iel_min
                ):  # the connected element number is 1-indexed
                    el.bcs[ibc, iface][3] += offset

        # also fix the index of the elements themselves in its their BC if relevant
        for el, ibc, iface in product(
            self.elem[iel_min:], range(self.nbc), range(2 * self.ndim)
        ):
            if int(el.bcs[ibc, iface][1]) > iel_min:
                el.bcs[ibc, iface][1] += offset



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import date

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata


sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------

project = "pymech"
_meta = metadata.metadata(project)
_today = date.today()

author = _meta.get("Author")
copyright = f"2016-{_today.year}"  # ", {author}"
master_doc = "index"

# The full version, including alpha/beta/rc tags
version = _meta.get("Version")
release = ".".join(version.split(".")[:3])

_py_min_version = _meta.get("Requires-Python").split(">=")[-1]

rst_prolog = f"""
.. |author| replace:: {author}
.. |today| replace:: {_today}
.. |py_min_version| replace:: {_py_min_version}
"""

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be

# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    # 'sphinx.ext.autosummary',
    # "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    #  "sphinx.ext.coverage",
    #  "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",  # Numpy-style docstrings
    #  "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst.md": "myst-nb",
}
nb_execution_mode = "cache"
nb_execution_in_temp = True
nb_execution_raise_on_error = True
nb_execution_show_tb = True
nb_execution_timeout = 600
nb_merge_streams = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "asv_bench/.asv"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# -- Other options -----------------------------------------------------------
#  autosummary_generate = True

autodoc_default_options = {
    "members": True,
}

autodoc_mock_imports = ["tvtk", "pymech._version"]

# -- Options for Intersphinx -------------------------------------------------

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    nek=("https://nek5000.github.io/NekDoc", None),
    xr=("https://xarray.pydata.org/en/stable/", None),
)

# -- MyST options ------------------------------------------------------------

myst_heading_anchors = 2
myst_enable_extensions = ["amsmath", "dollarmath", "colon_fence"]


def asv_publish():
    import sys
    from unittest.mock import patch

    cur_dir = os.getcwd()
    old_argv = sys.argv

    try:
        from asv.main import main

        os.chdir("asv_bench")
        sys.argv = ("asv", "publish")

        with patch("sys.exit"):
            main()
    finally:
        sys.argv = old_argv
        os.chdir(cur_dir)


asv_publish()



from __future__ import annotations

import struct
import sys

import numpy as np
from pymech.core import HexaData
from pymech.log import logger


def readrea(fname):
    """A function for reading .rea files for nek5000

    Parameters
    ----------
    fname : str
            file name
    """
    #
    try:
        infile = open(fname)
    except OSError as e:
        logger.critical(f"I/O error ({e.errno}): {e.strerror}")
        # return -1
    #
    # ---------------------------------------------------------------------------
    # count the number of boundary conditions
    # (it's too dangerous to infer it from the header)
    # ---------------------------------------------------------------------------
    #
    nbc = 0
    for line in infile:
        line_split = line.split()
        if "BOUNDARY" in line_split[2:] and "NO" not in line_split:
            nbc = nbc + 1

    infile.seek(0)
    #
    # ---------------------------------------------------------------------------
    # READ HEADER (2 lines) + ndim + number of parameters
    # ---------------------------------------------------------------------------
    #
    infile.readline()
    infile.readline()
    ndim = int(infile.readline().split()[0])
    npar = int(infile.readline().split()[0])
    #
    nface = 2 * ndim
    #
    # ---------------------------------------------------------------------------
    # READ parameters
    # ---------------------------------------------------------------------------
    #
    param = np.zeros((npar, 1))
    for ipar in range(npar):
        param[ipar] = float(infile.readline().split()[0])
    #
    # ---------------------------------------------------------------------------
    # skip passive scalars
    # ---------------------------------------------------------------------------
    #
    npscal_data = int(infile.readline().split()[0])
    for ipscal in range(npscal_data):
        infile.readline()
    #
    # ---------------------------------------------------------------------------
    # skip logical switches
    # ---------------------------------------------------------------------------
    #
    nswitch = int(infile.readline().split()[0])
    for iswitch in range(nswitch):
        infile.readline()
    #
    # ---------------------------------------------------------------------------
    # skip XFAC,YFAC,XZERO,YZERO
    # ---------------------------------------------------------------------------
    #
    infile.readline()
    #
    # ---------------------------------------------------------------------------
    # READ MESH
    # ---------------------------------------------------------------------------
    #
    infile.readline()
    nel = int(infile.readline().split()[0])
    #
    # initialize data structure
    lr1 = [2, 2, ndim - 1]
    var = [ndim, 0, 0, 0, 0]
    #
    data = HexaData(ndim, nel, lr1, var, nbc)
    #
    # read geometry
    for iel in range(nel):
        # skip element number and group
        infile.readline()
        for idim in range(var[0] - 1):  # if ndim == 3 do this twice
            for jdim in range(var[0]):
                fi = infile.readline().split()
                data.elem[iel].pos[jdim, idim, 0, 0] = float(fi[0])
                data.elem[iel].pos[jdim, idim, 0, 1] = float(fi[1])
                data.elem[iel].pos[jdim, idim, 1, 1] = float(fi[2])
                data.elem[iel].pos[jdim, idim, 1, 0] = float(fi[3])
    #
    # ---------------------------------------------------------------------------
    # CURVED SIDE DATA
    # ---------------------------------------------------------------------------
    #
    infile.readline()
    ncurved = int(infile.readline().split()[0])
    data.ncurv = ncurved
    for icurved in range(ncurved):
        line = infile.readline()
        if nel < 1e3:
            iedge = int(line[0:3]) - 1
            iel = int(line[3:6]) - 1
            data.elem[iel].curv[iedge][0] = float(line[6:20])
            data.elem[iel].curv[iedge][1] = float(line[20:34])
            data.elem[iel].curv[iedge][2] = float(line[34:48])
            data.elem[iel].curv[iedge][3] = float(line[48:62])
            data.elem[iel].curv[iedge][4] = float(line[62:76])
            data.elem[iel].ccurv[iedge] = line[76:79].split()[0]
        elif nel < 1e6:
            iedge = int(line[0:2]) - 1
            iel = int(line[2:8]) - 1
            data.elem[iel].curv[iedge][0] = float(line[8:22])
            data.elem[iel].curv[iedge][1] = float(line[22:36])
            data.elem[iel].curv[iedge][2] = float(line[36:50])
            data.elem[iel].curv[iedge][3] = float(line[50:64])
            data.elem[iel].curv[iedge][4] = float(line[64:78])
            data.elem[iel].ccurv[iedge] = line[78:81].split()[0]
        else:
            iedge = int(line[0:2]) - 1
            iel = int(line[2:12]) - 1
            data.elem[iel].curv[iedge][0] = float(line[12:26])
            data.elem[iel].curv[iedge][1] = float(line[26:40])
            data.elem[iel].curv[iedge][2] = float(line[40:54])
            data.elem[iel].curv[iedge][3] = float(line[54:68])
            data.elem[iel].curv[iedge][4] = float(line[68:82])
            data.elem[iel].ccurv[iedge] = line[82:85].split()[0]
    #
    # ---------------------------------------------------------------------------
    # BOUNDARY CONDITIONS
    # ---------------------------------------------------------------------------
    #
    infile.readline()  # ***** BOUNDARY CONDITIONS *****
    for ibc in range(nbc):
        infile.readline()  # ***** FLUID   BOUNDARY CONDITIONS ***** [or similar]
        for iel in range(nel):
            for iface in range(nface):
                line = infile.readline()
                if nel < 1e3:
                    data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
                    data.elem[iel].bcs[ibc, iface][1] = int(line[4:7])
                    data.elem[iel].bcs[ibc, iface][2] = int(line[7:10])
                    data.elem[iel].bcs[ibc, iface][3] = float(line[10:24])
                    data.elem[iel].bcs[ibc, iface][4] = float(line[24:38])
                    data.elem[iel].bcs[ibc, iface][5] = float(line[38:52])
                    data.elem[iel].bcs[ibc, iface][6] = float(line[52:66])
                    data.elem[iel].bcs[ibc, iface][7] = float(line[66:80])
                elif nel < 1e6:
                    data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
                    data.elem[iel].bcs[ibc, iface][1] = iel + 1
                    data.elem[iel].bcs[ibc, iface][2] = iface + 1
                    data.elem[iel].bcs[ibc, iface][3] = float(line[10:24])
                    data.elem[iel].bcs[ibc, iface][4] = float(line[24:38])
                    data.elem[iel].bcs[ibc, iface][5] = float(line[38:52])
                    data.elem[iel].bcs[ibc, iface][6] = float(line[52:66])
                    data.elem[iel].bcs[ibc, iface][7] = float(line[66:80])
                else:
                    data.elem[iel].bcs[ibc, iface][0] = line[1:3].strip()
                    data.elem[iel].bcs[ibc, iface][1] = int(line[4:15])
                    data.elem[iel].bcs[ibc, iface][2] = int(line[15:16])
                    data.elem[iel].bcs[ibc, iface][3] = float(line[16:34])
                    data.elem[iel].bcs[ibc, iface][4] = float(line[34:52])
                    data.elem[iel].bcs[ibc, iface][5] = float(line[52:70])
                    data.elem[iel].bcs[ibc, iface][6] = float(line[70:88])
                    data.elem[iel].bcs[ibc, iface][7] = float(line[88:106])
                # ignore some invalid internal 'E' conditions.
                # They are typically written this way by re2torea and Nek5000 ignores them.
                if (
                    data.elem[iel].bcs[ibc, iface][0] == "E"
                    and data.elem[iel].bcs[ibc, iface][3] == 0.0
                ):
                    data.elem[iel].bcs[ibc, iface][0] = ""
                    for j in range(1, 8):
                        data.elem[iel].bcs[ibc, iface][j] = 0
        ibc = ibc + 1
    #
    # ---------------------------------------------------------------------------
    # FORGET ABOUT WHAT FOLLOWS
    # ---------------------------------------------------------------------------
    #
    #
    # close file
    infile.close()
    #
    # output
    return data


# ==============================================================================
def writerea(fname, data):
    """A function for writing ascii .rea files for nek5000

    Parameters
    ----------
    fname : str
            file name
    data : :class:`pymech.core.HexaData`
            data structure
    """
    #
    try:
        outfile = open(fname, "w")
    except OSError as e:
        logger.critical(f"I/O error ({e.errno}): {e.strerror}")
        # return -1
    #
    # ---------------------------------------------------------------------------
    # READ HEADER (2 lines) + ndim + number of parameters
    # ---------------------------------------------------------------------------
    #
    outfile.write("****** PARAMETERS ******\n")
    outfile.write("   2.6000     NEKTON VERSION\n")
    outfile.write(f"   {data.ndim:1d} DIMENSIONAL RUN\n")
    outfile.write("         118 PARAMETERS FOLLOW\n")
    outfile.write("   1.00000     P001: DENSITY\n")
    outfile.write("  -1000.00     P002: VISCOSITY\n")
    outfile.write("   0.00000     P003: BETAG\n")
    outfile.write("   0.00000     P004: GTHETA\n")
    outfile.write("   0.00000     P005: PGRADX\n")
    outfile.write("   0.00000     P006: \n")
    outfile.write("   1.00000     P007: RHOCP\n")
    outfile.write("   1.00000     P008: CONDUCT\n")
    outfile.write("   0.00000     P009: \n")
    outfile.write("   0.00000     P010: FINTIME\n")
    outfile.write("   103         P011: NSTEPS\n")
    outfile.write("  -1.00000E-03 P012: DT\n")
    outfile.write("   0.00000     P013: IOCOMM\n")
    outfile.write("   0.00000     P014: IOTIME\n")
    outfile.write("   10          P015: IOSTEP\n")
    outfile.write("   0.00000     P016: PSSOLVER: 0=default\n")
    outfile.write("   1.00000     P017: \n")
    outfile.write("   0.00000     P018: GRID <0 --> # cells on screen\n")
    outfile.write("   0.00000     P019: INTYPE\n")
    outfile.write("   10.0000     P020: NORDER\n")
    outfile.write("   1.00000E-09 P021: DIVERGENCE\n")
    outfile.write("   1.00000E-09 P022: HELMHOLTZ\n")
    outfile.write("   0.00000     P023: NPSCAL\n")
    outfile.write("   1.00000E-02 P024: TOLREL\n")
    outfile.write("   1.00000E-02 P025: TOLABS\n")
    outfile.write("   1.00000     P026: COURANT/NTAU\n")
    outfile.write("   3.00000     P027: TORDER\n")
    outfile.write("   0.00000     P028: TORDER: mesh velocity (0: p28=p27)\n")
    outfile.write("   0.00000     P029: = magnetic visc if > 0, = -1/Rm if < 0\n")
    outfile.write("   0.00000     P030: > 0 ==> properties set in uservp()\n")
    outfile.write("   0.00000     P031: NPERT: #perturbation modes\n")
    outfile.write("   0.00000     P032: #BCs in re2 file, if > 0\n")
    outfile.write("   0.00000     P033: \n")
    outfile.write("   0.00000     P034: \n")
    outfile.write("   0.00000     P035: \n")
    outfile.write("   0.00000     P036: XMAGNET\n")
    outfile.write("   0.00000     P037: NGRIDS\n")
    outfile.write("   0.00000     P038: NORDER2\n")
    outfile.write("   0.00000     P039: NORDER3\n")
    outfile.write("   0.00000     P040: \n")
    outfile.write("   0.00000     P041: 1-->multiplicattive SEMG\n")
    outfile.write("   0.00000     P042: 0=gmres/1=pcg\n")
    outfile.write("   0.00000     P043: 0=semg/1=schwarz\n")
    outfile.write("   0.00000     P044: 0=E-based/1=A-based prec.\n")
    outfile.write("   0.00000     P045: Relaxation factor for DTFS\n")
    outfile.write("   0.00000     P046: reserved\n")
    outfile.write("   0.00000     P047: vnu: mesh material prop.\n")
    outfile.write("   0.00000     P048: \n")
    outfile.write("   0.00000     P049: \n")
    outfile.write("   0.00000     P050: \n")
    outfile.write("   0.00000     P051: \n")
    outfile.write("   0.00000     P052: IOHIS\n")
    outfile.write("   0.00000     P053: \n")
    outfile.write("   0.00000     P054: fixed flow rate dir: |p54|=1,2,3=x,y,z\n")
    outfile.write("   0.00000     P055: vol.flow rate (p54>0) or Ubar (p54<0)\n")
    outfile.write("   0.00000     P056: \n")
    outfile.write("   0.00000     P057: \n")
    outfile.write("   0.00000     P058: \n")
    outfile.write("   0.00000     P059: !=0 --> full Jac. eval. for each el.\n")
    outfile.write("   0.00000     P060: !=0 --> init. velocity to small nonzero\n")
    outfile.write("   0.00000     P061: \n")
    outfile.write("   0.00000     P062: >0 --> force byte_swap for output\n")
    outfile.write("   8.00000     P063: =8 --> force 8-byte output\n")
    outfile.write("   0.00000     P064: =1 --> perturbation restart\n")
    outfile.write("   1.00000     P065: #iofiles (eg, 0 or 64); <0 --> sep. dirs\n")
    outfile.write("   6.00000     P066: output : <0=ascii, else binary\n")
    outfile.write("   6.00000     P067: restart: <0=ascii, else binary\n")
    outfile.write("   0.00000     P068: iastep: freq for avg_all (0=iostep)\n")
    outfile.write("   0.00000     P069:       : freq of srf dump\n")
    outfile.write("   0.00000     P070: \n")
    outfile.write("   0.00000     P071: \n")
    outfile.write("   0.00000     P072: \n")
    outfile.write("   0.00000     P073: \n")
    outfile.write("   0.00000     P074: \n")
    outfile.write("   0.00000     P075: \n")
    outfile.write("   0.00000     P076: \n")
    outfile.write("   0.00000     P077: \n")
    outfile.write("   0.00000     P078: \n")
    outfile.write("   0.00000     P079: \n")
    outfile.write("   0.00000     P080: \n")
    outfile.write("   0.00000     P081: \n")
    outfile.write("   0.00000     P082: \n")
    outfile.write("   0.00000     P083: \n")
    outfile.write("   0.00000     P084: != 0 --> sets initial timestep if p12>0\n")
    outfile.write("   0.00000     P085: dt retio of p84 !=0, for timesteps>0\n")
    outfile.write("   0.00000     P086: reserved\n")
    outfile.write("   0.00000     P087: \n")
    outfile.write("   0.00000     P088: \n")
    outfile.write("   0.00000     P089: \n")
    outfile.write("   0.00000     P090: \n")
    outfile.write("   0.00000     P091: \n")
    outfile.write("   0.00000     P092: \n")
    outfile.write("   20.0000     P093: Number of previous pressure solns saved\n")
    outfile.write("   9.00000     P094: start projecting velocity after p94 step\n")
    outfile.write("   9.00000     P095: start projecting pressure after p95 step\n")
    outfile.write("   0.00000     P096: \n")
    outfile.write("   0.00000     P097: \n")
    outfile.write("   0.00000     P098: \n")
    outfile.write("   3.00000     P099: dealiasing: <0--> off /3--> old /4-->new\n")
    outfile.write("   0.00000     P100: \n")
    outfile.write("   0.00000     P101: Number of additional modes to filter\n")
    outfile.write("   0.00000     P102: Dump out divergence at each time step\n")
    outfile.write("   0.01000     P103: weight of stabilizing filter\n")
    outfile.write("   0.00000     P104: \n")
    outfile.write("   0.00000     P105: \n")
    outfile.write("   0.00000     P106: \n")
    outfile.write("   0.00000     P107: !=0 --> add h2 array in hmholtz eqn\n")
    outfile.write("   0.00000     P108: \n")
    outfile.write("   0.00000     P109: \n")
    outfile.write("   0.00000     P110: \n")
    outfile.write("   0.00000     P111: \n")
    outfile.write("   0.00000     P112: \n")
    outfile.write("   0.00000     P113: \n")
    outfile.write("   0.00000     P114: \n")
    outfile.write("   0.00000     P115: \n")
    outfile.write("   0.00000     P116: \n")
    outfile.write("   0.00000     P117: \n")
    outfile.write("   0.00000     P118: \n")
    outfile.write("      4  Lines of passive scalar data follows2 CONDUCT, 2RHOCP\n")
    outfile.write(
        "   1.00000        1.00000        1.00000        1.00000        1.00000\n"
    )
    outfile.write("   1.00000        1.00000        1.00000        1.00000\n")
    outfile.write(
        "   1.00000        1.00000        1.00000        1.00000        1.00000\n"
    )
    outfile.write("   1.00000        1.00000        1.00000        1.00000\n")
    outfile.write("         13   LOGICAL SWITCHES FOLLOW\n")
    outfile.write(" T      IFFLOW\n")
    outfile.write(" F      IFHEAT\n")
    outfile.write(" T      IFTRAN\n")
    outfile.write(
        " T F F F F F F F F F F  IFNAV & IFADVC (convection in P.S. fields)\n"
    )
    outfile.write(
        " F F T T T T T T T T T T  IFTMSH (IF mesh for this field is T mesh)\n"
    )
    outfile.write(" F      IFAXIS\n")
    outfile.write(" F      IFSTRS\n")
    outfile.write(" F      IFSPLIT\n")
    outfile.write(" F      IFMGRID\n")
    outfile.write(" F      IFMODEL\n")
    outfile.write(" F      IFKEPS\n")
    outfile.write(" F      IFMVBD\n")
    outfile.write(" F      IFCHAR\n")
    outfile.write(
        "   2.00000       2.00000      -1.00000      -1.00000     XFAC,YFAC,XZERO,YZERO\n"
    )
    #
    # vertex data
    outfile.write(
        "  ***** MESH DATA *****  6 lines are X,Y,Z;X,Y,Z. Columns corners 1-4;5-8\n"
    )
    outfile.write(f"  {data.nel:10d} {data.ndim:1d} {data.nel:10d} NEL,NDIM,NELV\n")
    for iel in range(data.nel):
        outfile.write("           ELEMENT {:10d} [  1a]    GROUP 1\n".format(iel + 1))
        if data.ndim == 2:
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[0, 0, 0, 0],
                    data.elem[iel].pos[0, 0, 0, -1],
                    data.elem[iel].pos[0, 0, -1, -1],
                    data.elem[iel].pos[0, 0, -1, 0],
                )
            )
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[1, 0, 0, 0],
                    data.elem[iel].pos[1, 0, 0, -1],
                    data.elem[iel].pos[1, 0, -1, -1],
                    data.elem[iel].pos[1, 0, -1, 0],
                )
            )
        elif data.ndim == 3:
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[0, 0, 0, 0],
                    data.elem[iel].pos[0, 0, 0, -1],
                    data.elem[iel].pos[0, 0, -1, -1],
                    data.elem[iel].pos[0, 0, -1, 0],
                )
            )
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[1, 0, 0, 0],
                    data.elem[iel].pos[1, 0, 0, -1],
                    data.elem[iel].pos[1, 0, -1, -1],
                    data.elem[iel].pos[1, 0, -1, 0],
                )
            )
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[2, 0, 0, 0],
                    data.elem[iel].pos[2, 0, 0, -1],
                    data.elem[iel].pos[2, 0, -1, -1],
                    data.elem[iel].pos[2, 0, -1, 0],
                )
            )
            #
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[0, -1, 0, 0],
                    data.elem[iel].pos[0, -1, 0, -1],
                    data.elem[iel].pos[0, -1, -1, -1],
                    data.elem[iel].pos[0, -1, -1, 0],
                )
            )
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[1, -1, 0, 0],
                    data.elem[iel].pos[1, -1, 0, -1],
                    data.elem[iel].pos[1, -1, -1, -1],
                    data.elem[iel].pos[1, -1, -1, 0],
                )
            )
            outfile.write(
                "{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                    data.elem[iel].pos[2, -1, 0, 0],
                    data.elem[iel].pos[2, -1, 0, -1],
                    data.elem[iel].pos[2, -1, -1, -1],
                    data.elem[iel].pos[2, -1, -1, 0],
                )
            )
    #
    # curved side data
    outfile.write("  ***** CURVED SIDE DATA *****\n")
    outfile.write(
        f"  {data.ncurv:10d} Curved sides follow IEDGE,IEL,CURVE(I),I=1,5, CCURVE\n"
    )
    for iel in range(data.nel):
        if data.nel < 1e3:
            #
            for iedge in range(12):
                if data.elem[iel].ccurv[iedge] != "":
                    outfile.write(
                        "{:3d}{:3d}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:>2s}\n".format(
                            iedge + 1,
                            iel + 1,
                            data.elem[iel].curv[iedge][0],
                            data.elem[iel].curv[iedge][1],
                            data.elem[iel].curv[iedge][2],
                            data.elem[iel].curv[iedge][3],
                            data.elem[iel].curv[iedge][4],
                            data.elem[iel].ccurv[iedge],
                        )
                    )
        elif data.nel < 1e6:
            #
            for iedge in range(12):
                if data.elem[iel].ccurv[iedge] != "":
                    outfile.write(
                        "{:2d}{:6d}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:>2s}\n".format(
                            iedge + 1,
                            iel + 1,
                            data.elem[iel].curv[iedge][0],
                            data.elem[iel].curv[iedge][1],
                            data.elem[iel].curv[iedge][2],
                            data.elem[iel].curv[iedge][3],
                            data.elem[iel].curv[iedge][4],
                            data.elem[iel].ccurv[iedge],
                        )
                    )
        else:
            #
            for iedge in range(12):
                if data.elem[iel].ccurv[iedge] != "":
                    outfile.write(
                        "{:2d}{:10d}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:>2s}\n".format(
                            iedge + 1,
                            iel + 1,
                            data.elem[iel].curv[iedge][0],
                            data.elem[iel].curv[iedge][1],
                            data.elem[iel].curv[iedge][2],
                            data.elem[iel].curv[iedge][3],
                            data.elem[iel].curv[iedge][4],
                            data.elem[iel].ccurv[iedge],
                        )
                    )
    #
    # boundary conditions data
    outfile.write("  ***** BOUNDARY CONDITIONS *****\n")
    outfile.write("  ***** FLUID BOUNDARY CONDITIONS *****\n")
    for ibc in range(data.nbc):
        if ibc == 1:
            outfile.write("  ***** THERMAL BOUNDARY CONDITIONS *****\n")
        elif ibc > 1:
            outfile.write(
                "  ***** PASSIVE SCALAR {:4d} BOUNDARY CONDITIONS *****\n".format(
                    ibc - 1
                )
            )
        for iel in range(data.nel):
            for iface in range(2 * data.ndim):
                # if no boundary condition is specified, write 'E' iel iface [...]
                # this is the behaviour of re2torea.
                if data.elem[iel].bcs[ibc, iface][0] == "":
                    data.elem[iel].bcs[ibc, iface][0] = "E"
                    data.elem[iel].bcs[ibc, iface][1] = iel + 1
                    data.elem[iel].bcs[ibc, iface][2] = iface + 1
                if data.nel < 1e3:
                    outfile.write(
                        " {:2s} {:3d}{:3d}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                            data.elem[iel].bcs[ibc, iface][0],
                            data.elem[iel].bcs[ibc, iface][1],
                            data.elem[iel].bcs[ibc, iface][2],
                            data.elem[iel].bcs[ibc, iface][3],
                            data.elem[iel].bcs[ibc, iface][4],
                            data.elem[iel].bcs[ibc, iface][5],
                            data.elem[iel].bcs[ibc, iface][6],
                            data.elem[iel].bcs[ibc, iface][7],
                        )
                    )
                elif data.nel < 1e6:
                    outfile.write(
                        " {:2s} {:6d}{:14.6e}{:14.6e}{:14.6e}{:14.6e}{:14.6e}\n".format(
                            data.elem[iel].bcs[ibc, iface][0],
                            data.elem[iel].bcs[ibc, iface][1],
                            data.elem[iel].bcs[ibc, iface][3],
                            data.elem[iel].bcs[ibc, iface][4],
                            data.elem[iel].bcs[ibc, iface][5],
                            data.elem[iel].bcs[ibc, iface][6],
                            data.elem[iel].bcs[ibc, iface][7],
                        )
                    )
                else:
                    outfile.write(
                        " {:2s} {:11d}{:1d}{:18.11e}{:18.11e}{:18.11e}{:18.11e}{:18.11e}\n".format(
                            data.elem[iel].bcs[ibc, iface][0],
                            data.elem[iel].bcs[ibc, iface][1],
                            data.elem[iel].bcs[ibc, iface][2],
                            data.elem[iel].bcs[ibc, iface][3],
                            data.elem[iel].bcs[ibc, iface][4],
                            data.elem[iel].bcs[ibc, iface][5],
                            data.elem[iel].bcs[ibc, iface][6],
                            data.elem[iel].bcs[ibc, iface][7],
                        )
                    )

    if data.nbc < 2:
        outfile.write("  ***** NO THERMAL BOUNDARY CONDITIONS *****\n")
    outfile.write("    0 PRESOLVE/RESTART OPTIONS  *****\n")
    outfile.write("    7         INITIAL CONDITIONS *****\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write(" C Default\n")
    outfile.write("  ***** DRIVE FORCE DATA ***** BODY FORCE, FLOW, Q\n")
    outfile.write("    4                 Lines of Drive force data follow\n")
    outfile.write(" C\n")
    outfile.write(" C\n")
    outfile.write(" C\n")
    outfile.write(" C\n")
    outfile.write(" ***** Variable Property Data ***** Overrrides Parameter data.\n")
    outfile.write("    1 Lines follow.\n")
    outfile.write("    0 PACKETS OF DATA FOLLOW\n")
    outfile.write(" ***** HISTORY AND INTEGRAL DATA *****\n")
    outfile.write("    0   POINTS.  Hcode, I,J,H,IEL\n")
    outfile.write(" ***** OUTPUT FIELD SPECIFICATION *****\n")
    outfile.write("    6 SPECIFICATIONS FOLLOW\n")
    outfile.write("    F      COORDINATES\n")
    outfile.write("    T      VELOCITY\n")
    outfile.write("    T      PRESSURE\n")
    outfile.write("    F      TEMPERATURE\n")
    outfile.write("    F      TEMPERATURE GRADIENT\n")
    outfile.write("    0      PASSIVE SCALARS\n")
    outfile.write(" ***** OBJECT SPECIFICATION *****\n")
    outfile.write("        0 Surface Objects\n")
    outfile.write("        0 Volume  Objects\n")
    outfile.write("        0 Edge    Objects\n")
    outfile.write("        0 Point   Objects\n")
    #
    # close file
    outfile.close()
    #
    # output
    return 0


def readre2(fname):
    """A function for reading .re2 files for nek5000

    Parameters
    ----------
    fname : str
            file name
    """
    #
    try:
        infile = open(fname, "rb")
    except OSError as e:
        logger.critical(f"I/O error ({e.errno}): {e.strerror}")
        return -1
    # the header for re2 files is 80 ASCII bytes, something like
    # #v002    18669  2    18669 this is the hdr                                      %
    header = infile.read(80).split()
    nel = int(header[1])
    ndim = int(header[2])
    # always double precision
    wdsz = 8
    realtype = "d"

    # detect endianness
    etagb = infile.read(4)
    etagL = struct.unpack("<f", etagb)[0]
    etagL = int(etagL * 1e5) / 1e5
    etagB = struct.unpack(">f", etagb)[0]
    etagB = int(etagB * 1e5) / 1e5
    if etagL == 6.54321:
        logger.debug("Reading little-endian file\n")
        emode = "<"
        endian = "little"
    elif etagB == 6.54321:
        logger.debug("Reading big-endian file\n")
        emode = ">"
        endian = "big"
    else:
        logger.error("Could not interpret endianness")
        return -3
    #
    # there are no GLL points here, only quad/hex vertices
    lr1 = [2, 2, ndim - 1]
    npel = 2**ndim
    # the file only contains geometry
    var = [ndim, 0, 0, 0, 0]
    # allocate structure
    data = HexaData(ndim, nel, lr1, var, 1)
    #
    # some metadata
    data.wdsz = wdsz
    data.endian = endian
    #
    # read the whole geometry into a buffer
    # for some reason each element is prefixed with 8 bytes of zeros, it's not clear to me why.
    # This is the reason for the +1 here, then the first number is ignored.
    buf = infile.read((ndim * npel + 1) * wdsz * nel)
    # elem_shape = [ndim, ndim-1, 2, 2]  # nvar, lz, ly, lx
    for (iel, el) in enumerate(data.elem):
        fi = np.frombuffer(
            buf,
            dtype=emode + realtype,
            count=ndim * npel + 1,
            offset=(ndim * npel + 1) * wdsz * iel,
        )
        # the data is stored in the following order (2D|3D):
        # x1, x2, x4, x3; | x5, x6, x8, x7;
        # y1, y2, y4, y3; | y5, y6, y8, y7;
        # ----------------
        # z1, z2, z4, z3; z5, z6, z8, z7;
        # where 1-8 is the ordering of the points in memory
        for idim in range(ndim):  # x, y, [z]
            for iz in range(
                ndim - 1
            ):  # this does only one iteration in 2D, and in 3D does one iteration for 1-4 and one for 5-8
                el.pos[idim, iz, 0, 0] = fi[npel * idim + 4 * iz + 1]
                el.pos[idim, iz, 0, 1] = fi[npel * idim + 4 * iz + 2]
                el.pos[idim, iz, 1, 1] = fi[npel * idim + 4 * iz + 3]
                el.pos[idim, iz, 1, 0] = fi[npel * idim + 4 * iz + 4]
    #
    # read curved sides
    # the number of curved sides is stored as a double,
    # then each curved side is 64 bytes:
    # iel iface p1 p2 p3 p4 p5 ctype
    # where p1-5 are f64 parameters and ctype is the type of curvature in ASCII
    ncparam = 8
    buf = infile.read(wdsz)
    ncurv = int(np.frombuffer(buf)[0])
    logger.debug(f"Found {ncurv} curved sides")
    data.ncurv = ncurv
    buf = infile.read(wdsz * (ncparam * ncurv))
    for icurv in range(ncurv):
        # interpret the data
        curv = np.frombuffer(
            buf,
            dtype=emode + realtype,
            count=ncparam,
            offset=icurv * ncparam * wdsz,
        )
        iel = int(curv[0]) - 1
        iedge = int(curv[1]) - 1
        cparams = curv[2:7]
        # select only the first byte, because it turns out the later bytes may contain garbage.
        # typically, it's b'C\x00\x00\x00\x00\x00\x00\x00' or b'C\x00\x00\x00\x00\x00\xe0?'.
        # AFAIK the curvature types are always one character long anyway.
        ctype = curv[7].tobytes()[:1].decode("utf-8")
        # fill in the data structure
        data.elem[iel].curv[iedge, :] = cparams
        data.elem[iel].ccurv[iedge] = ctype
    #
    # read boundary conditions
    # there can be more than one field, and we won't know until we'vre reached the end
    nbcparam = 8
    buf = infile.read(wdsz)
    ifield = 0
    while buf != b"":
        # the data is initialized with one BC field, we might need to allocate another
        if ifield > 0:
            data.nbc = data.nbc + 1
            for el in data.elem:
                empty_bcs = np.zeros(el.bcs[:1, :].shape, dtype=el.bcs.dtype)
                el.bcs = np.concatenate((el.bcs, empty_bcs))
        nbclines = int(np.frombuffer(buf)[0])
        logger.debug(
            f"Found {nbclines} external boundary conditions for field {ifield}"
        )
        buf = infile.read(wdsz * (nbcparam * nbclines))
        for ibc in range(nbclines):
            # interpret the data
            bc = np.frombuffer(
                buf,
                dtype=emode + realtype,
                count=nbcparam,
                offset=ibc * nbcparam * wdsz,
            )
            iel = int(bc[0]) - 1
            iface = int(bc[1]) - 1
            bcparams = bc[2:7]
            bctype = bc[7].tobytes().decode("utf-8").rstrip()  # remove trailing spaces
            # fill in the data structure
            data.elem[iel].bcs[ifield, iface][0] = bctype
            data.elem[iel].bcs[ifield, iface][1] = iel + 1
            data.elem[iel].bcs[ifield, iface][2] = iface + 1
            for ipar in range(5):
                data.elem[iel].bcs[ifield, iface][3 + ipar] = bcparams[ipar]
        ifield = ifield + 1
        # try reading the number of conditions in the next field
        buf = infile.read(wdsz)
    infile.close()
    return data


def writere2(fname, data):
    """A function for writing binary .re2 files for nek5000

    Parameters
    ----------
    fname : str
            file name
    data : :class:`pymech.core.HexaData`
            data structure
    """
    #
    # ---------------------------------------------------------------------------
    # CHECK INPUT DATA
    # ---------------------------------------------------------------------------
    #
    # We could extract the corners, but for now just return an error if lr1 is too large
    if data.lr1 != [2, 2, data.ndim - 1]:
        logger.critical(
            "wrong element dimensions for re2 file! {} != {}".format(
                data.lr1, [2, 2, data.ndim - 1]
            )
        )
        return -2
    #
    if data.var[0] != data.ndim:
        logger.critical(
            "wrong number of geometric variables for re2 file! expected {}, found {}".format(
                data.ndim, data.var[0]
            )
        )
        return -3
    #
    # Open file
    try:
        outfile = open(fname, "wb")
    except OSError as e:
        logger.critical(f"I/O error ({e.errno}): {e.strerror}")
        return -1
    #
    # ---------------------------------------------------------------------------
    # WRITE HEADER
    # ---------------------------------------------------------------------------
    #
    # always double precision
    #  wdsz = 8
    #  realtype = 'd'
    nel = data.nel
    ndim = data.ndim
    header = f"#v002{nel:9d}{ndim:3d}{nel:9d} this is the hdr"
    header = header.ljust(80)
    outfile.write(header.encode("utf-8"))
    #
    # decide endianness
    if data.endian in ("big", "little"):
        byteswap = data.endian != sys.byteorder
        logger.debug(f"Writing {data.endian}-endian file")
    else:
        byteswap = False
        logger.warning(
            f"Unrecognized endianness {data.endian}, "
            f"writing native {sys.byteorder}-endian file"
        )

    def correct_endianness(a):
        """Return the array with the requested endianness"""
        if byteswap:
            return a.byteswap()
        else:
            return a

    #
    # write tag (to specify endianness)
    endianbytes = np.array([6.54321], dtype=np.float32)
    correct_endianness(endianbytes).tofile(outfile)
    #
    # ---------------------------------------------------------------------------
    # WRITE DATA
    # ---------------------------------------------------------------------------
    #
    # compute total number of points per element
    npel = 2**ndim

    def write_data_to_file(a):
        """Write the geometry of an element to the output file in double precision"""
        correct_endianness(a).tofile(outfile)

    #
    # write geometry (adding eight bytes of zeros before each element)
    xyz = np.zeros(
        (npel * ndim + 1,)
    )  # array storing reordered geometry data (with a zero in the first position)
    for el in data.elem:
        # the data is stored in the following order (2D|3D):
        # x1, x2, x4, x3; | x5, x6, x8, x7;
        # y1, y2, y4, y3; | y5, y6, y8, y7;
        # ----------------
        # z1, z2, z4, z3; z5, z6, z8, z7;
        # where 1-8 is the ordering of the points in memory
        for idim in range(ndim):  # x, y, [z]
            for iz in range(
                ndim - 1
            ):  # this does only one iteration in 2D, and in 3D does one iteration for 1-4 and one for 5-8
                xyz[npel * idim + 4 * iz + 1] = el.pos[idim, iz, 0, 0]
                xyz[npel * idim + 4 * iz + 2] = el.pos[idim, iz, 0, 1]
                xyz[npel * idim + 4 * iz + 3] = el.pos[idim, iz, 1, 1]
                xyz[npel * idim + 4 * iz + 4] = el.pos[idim, iz, 1, 0]
        write_data_to_file(xyz)
    #
    # write curve sides data
    # locate curved edges
    curved_edges = []
    for (iel, el) in enumerate(data.elem):
        for iedge in range(12):
            if el.ccurv[iedge] != "":
                curved_edges.append((iel, iedge))
    # write number of curved edges
    ncurv = len(curved_edges)
    if ncurv != data.ncurv:
        logger.warning(
            f"wrong number of curved edges: expected {data.ncurv}, found {ncurv}"
        )
    ncurvf = np.array([ncurv], dtype=np.float64)
    write_data_to_file(ncurvf)
    # format curve data
    cdata = np.zeros((ncurv,), dtype="f8, f8, f8, f8, f8, f8, f8, S8")
    for (cdat, (iel, iedge)) in zip(cdata, curved_edges):
        el = data.elem[iel]
        cdat[0] = iel + 1
        cdat[1] = iedge + 1
        # curve parameters
        for j in range(5):
            cdat[2 + j] = el.curv[iedge, j]
        # encode the string as a byte array padded with spaces
        cdat[7] = el.ccurv[iedge].encode("utf-8")
    # write to file
    write_data_to_file(cdata)
    #
    # write boundary conditions for each field
    for ifield in range(data.nbc):
        # locate faces with boundary conditions
        bc_faces = []
        for (iel, el) in enumerate(data.elem):
            for iface in range(2 * ndim):
                bctype = el.bcs[ifield, iface][0]
                # internal boundary conditions are not written to .re2 files by reatore2
                # and are apparently ignored by Nek5000 even in .rea files
                if bctype != "" and bctype != "E":
                    bc_faces.append((iel, iface))
        nbcs = len(bc_faces)
        nbcsf = np.array([nbcs], dtype=np.float64)
        write_data_to_file(nbcsf)
        # initialize and format data
        bcdata = np.zeros((nbcs,), dtype="f8, f8, f8, f8, f8, f8, f8, S8")
        for (bc, (iel, iface)) in zip(bcdata, bc_faces):
            el = data.elem[iel]
            bc[0] = iel + 1
            bc[1] = iface + 1
            for j in range(5):
                bc[2 + j] = el.bcs[ifield, iface][3 + j]
            # encode the string as a byte array padded with spaces
            bc[7] = el.bcs[ifield, iface][0].encode("utf-8").ljust(8)
        # write to file
        write_data_to_file(bcdata)
    #
    # close file
    outfile.close()
    # rerurn
    return 0



"""Module for converting :class:`pymech.core.HexaData` objects to vtk"""
import os
from itertools import product
from pathlib import Path
from warnings import warn

import numpy as np

from .log import logger

logger.warning(
    "The module pymech.vtksuite is experimental in nature and "
    "may have some rough edges. The functions can also change in the future."
)

try:
    from tvtk.api import tvtk, write_data
except ImportError:
    logger.warning("To use VTK functions,\n    pip install mayavi")


__all__ = ("exa2vtk", "writevtk")


# ==============================================================================
def exa2vtk(field, downsample=False):
    warn(
        (
            "Function pymech.vtksuite.exa2vtk is now pymech.vtksuite.hexa2vtk. "
            "This module is kept for backwards "
            "compatibility and would disappear in version 2.0.0"
        ),
        DeprecationWarning,
    )
    return hexa2vtk(field, downsample)


def hexa2vtk(field, downsample=False):
    """A function for converting :class:`pymech.core.HexaData` to `Traited VTK`_ dataset. The
    returned dataset can be manipulated with libraries which accept a VTK
    object, for example Mayavi_.

    .. _Traited VTK: https://docs.enthought.com/mayavi/tvtk/README.html

    .. todo::

        Try https://github.com/pyvista/pyvista-xarray

    Example
    -------
    This also requires you to have a GUI toolkit installed: either PyQt4,
    PySide, PySide2, PyQt5 or wxPython.

    .. code-block:: python

       import pymech as pm
       from pymech.vtksuite import exa2vtk
       from mayavi import mlab

       field = pm.readnek("tests/nek/channel3D_0.f00001")
       dataset = exa2vtk(field)
       mlab.pipeline.add_dataset(dataset)

    Instead of MayaVi_ you could use also use something high-level like PyVista_
    to wrap the underlying VTK object and later visualize them.

    .. code-block:: python

        import pyvista as pv
        dataset = pv.wrap(dataset._vtk_obj)
        dataset.plot()

    .. _MayaVi: https://docs.enthought.com/mayavi/mayavi/mlab.html
    .. _PyVista: https://docs.pyvista.org/getting-started/index.html

    Parameters
    ----------
    field : :class:`pymech.core.HexaData`
            a dataset in nekdata format
    downsample : bool
            flag T/F

    Returns
    -------
    dataset : tvtk.tvtk_classes.unstructured_grid.UnstructuredGrid
            a VTK dataset
    """
    #
    if downsample:
        ixs = field.lr1[0] - 1
        iys = field.lr1[1] - 1
        izs = max(field.lr1[2] - 1, 1)
    else:
        ixs = 1
        iys = 1
        izs = 1
    #
    iix = range(0, field.lr1[0], ixs)
    nix = len(iix)
    iiy = range(0, field.lr1[1], iys)
    niy = len(iiy)
    iiz = range(0, field.lr1[2], izs)
    niz = len(iiz)
    #
    nppel = nix * niy * niz
    nepel = (nix - 1) * (niy - 1) * max((niz - 1), 1)
    nel = field.nel * nepel
    #
    if field.ndim == 3:
        nvert = 8
        cellType = tvtk.Hexahedron().cell_type
    else:
        nvert = 4
        cellType = tvtk.Quad().cell_type
    #
    ct = np.array(nel * [cellType])
    of = np.arange(0, nvert * nel, nvert)

    ce = np.zeros(nel * (nvert + 1))
    ce[np.arange(nel) * (nvert + 1)] = nvert

    if field.var[0] != 0:
        r = np.zeros((nvert * nel, 3))
    if field.var[1] != 0:
        v = np.zeros((nvert * nel, 3))
    if field.var[2] == 1:
        p = np.zeros(nvert * nel)
    if field.var[3] == 1:
        T = np.zeros(nvert * nel)
    if field.var[4] != 0:
        S = np.zeros((nvert * nel, field.var[4]))
    #
    ice = -(nvert + 1)

    for iel in range(field.nel):
        for (iz, ez), (iy, ey), (ix, ex) in product(
            enumerate(iiz), enumerate(iiy), enumerate(iix)
        ):
            iarray = iel * nppel + ix + iy * nix + iz * (nix * niy)

            # Downsample copy into a column vector
            if field.var[0] == 3:
                r[iarray, :] = field.elem[iel].pos[:, ez, ey, ex]
            if field.var[1] == 3:
                v[iarray, :] = field.elem[iel].vel[:, ez, ey, ex]
            if field.var[2] == 1:
                p[iarray] = field.elem[iel].pres[:, ez, ey, ex]
            if field.var[3] == 1:
                T[iarray] = field.elem[iel].temp[:, ez, ey, ex]
            if field.var[4] != 0:
                S[iarray, :] = field.elem[iel].scal[:, ez, ey, ex]
        if field.var[0] == 3:
            for iz, iy, ix in product(
                range(max(niz - 1, 1)), range(niy - 1), range(nix - 1)
            ):
                ice = ice + nvert + 1
                for face in range(field.ndim - 1):
                    cell_id = iel * nppel + ix + iy * nix + (iz + face) * nix * niy

                    ce[ice + face * 4 + 1] = cell_id
                    ce[ice + face * 4 + 2] = cell_id + 1
                    ce[ice + face * 4 + 3] = cell_id + nix + 1
                    ce[ice + face * 4 + 4] = cell_id + nix

    # create the array of cells
    ca = tvtk.CellArray()
    ca.set_cells(nel, ce)
    # create the unstructured dataset
    dataset = tvtk.UnstructuredGrid(points=r)
    # set the cell types
    dataset.set_cells(ct, of, ca)
    # set the data
    idata = 0
    if field.var[1] != 0:
        dataset.point_data.vectors = v
        dataset.point_data.vectors.name = "vel"
        idata += 1
    if field.var[2] == 1:
        dataset.point_data.scalars = p
        dataset.point_data.scalars.name = "pres"
        idata += 1
    if field.var[3] == 1:
        dataset.point_data.add_array(T)
        dataset.point_data.get_array(idata).name = "temp"
        idata += 1
    if field.var[4] != 0:
        for ii in range(field.var[4]):
            dataset.point_data.add_array(S[:, ii])
            dataset.point_data.get_array(ii + idata).name = "scal_%d" % (ii + 1)
    #
    dataset.point_data.update()
    #
    return dataset


def writevtk(fname, data):
    """A function for writing binary data in the XML VTK format

    Parameters
    ----------
    fname : str
            file name
    data : :class:`pymech.core.HexaData`
            data organised after reading a file

    """
    ext = ".vtp"
    fname = Path(fname)
    if fname.suffix != ext:
        logger.info(f"Renaming {fname} with extension .vtp")
        fname = fname.with_suffix(ext)

    vtk_dataset = hexa2vtk(data)
    write_data(vtk_dataset, os.fspath(fname))



"""Dummy module for compatibility with pymech version < 1.5.0


.. class:: pymech.exadata.datalims

   Use :class:`pymech.core.DataLims` instead

.. class:: pymech.exadata.elem

   Use :class:`pymech.core.Elem` instead

.. class:: pymech.exadata.exadata

   Use :class:`pymech.core.HexaData` instead

"""
from warnings import warn

from pymech.core import DataLims as datalims  # noqa
from pymech.core import Elem as elem  # noqa
from pymech.core import HexaData as exadata  # noqa


warn(
    (
        "Module pymech.exadata is now pymech.core. This module is kept for backwards "
        "compatibility and would disappear in version 2.0.0"
    ),
    DeprecationWarning,
)
